// Equinox Desktop Environment - main panel
// Copyright (C) 2000-2002 Martin Pekar
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "workpanel.h"
#include "aboutdialog.h"

#include "panelbutton.h"
#include "taskbutton.h"
#include "cpumonitor.h"
#include "dock.h"
#include "mainmenu.h"

#include <efltk/Fl_WM.h>
#include <efltk/Fl_String.h>
#include <efltk/Fl_Pack.h>
#include <efltk/Fl_Divider.h>
#include <efltk/Fl_Input_Browser.h>

#include "item.h"
#include <errno.h>

static Fl_Image desktop_pix((const char **)desktop_xpm);
static Fl_Image keyboard_pix((const char **)keyboard_xpm);
static Fl_Image sound_pix((const char **)sound_xpm);
static Fl_Image run_pix((const char **)run_xpm);

class Fl_Update_Window : public Fl_Window
{
public:
    Fl_Update_Window(int, int, int, int, char*);
    void setAutoHide(int a) { autoHide = a; }

    int handle(int event);

    int autoHide;
    int initX, initY, initW, initH;
};

Fl_Update_Window *mPanelWindow;

Fl_Button 	 *mClockBox;
Fl_Group 	 *mModemLeds;
Fl_Box 		 *mLedIn, *mLedOut;

Fl_Input_Browser *runBrowser;

PanelButton 	 *mProgramsMenu, *mKbdSelect;
PanelButton 	 *mWorkspace;

Dock    *dock;
TaskBar *tasks;

Fl_Config pGlobalConfig(fl_find_config_file("ede.conf", true));

void 		 getKeyboard(PanelButton *);
void 		 updateStats(void *);

static EItemGroup *mBrowser;

/////////////////////////////////////////////////////////////////////////////

Fl_Update_Window::Fl_Update_Window(int X, int Y, int W, int H, char *label=0)
    : Fl_Window(X, Y, W, H, label)
{
    autoHide = 0;
    initX = X;
    initY = Y;
    initW = W;
    initH = H;
}

int Fl_Update_Window::handle(int event) 
{
    switch (event) {
    case FL_SHOW: {
        int ret = Fl_Window::handle(event);
        // h() pixel space at bottom of the screen
        if(shown()) Fl_WM::set_window_strut(fl_xid(mPanelWindow), 0, 0, 0, h());
        return ret;
    }
    case FL_ENTER:
        position(initX, initY);
        if(shown()) Fl_WM::set_window_strut(fl_xid(mPanelWindow), 0, 0, 0, h());
        break;
    case FL_LEAVE:
        if (autoHide) {
            position(initX, initY+h()-4);
            if(shown()) Fl_WM::set_window_strut(fl_xid(mPanelWindow), 0, 0, 0, 4);
        }
        break;
    }
    Fl_Window::handle(event);
    return 1;
}

void setWorkspace(Fl_Button *, void *w)
{
    Fl_WM::set_current_workspace((int) w);
}

void setKeyboard(const Fl_String &pKbdLayout)
{
    if(!pKbdLayout.empty()) {
        Fl_String ApplyString("setxkbmap " + pKbdLayout + " &");
        fl_start_child_process(ApplyString);

        Fl_String pShortKbd = pKbdLayout.sub_str(0, 2);
        mKbdSelect->tooltip(pKbdLayout);
        mKbdSelect->label(pShortKbd);
        mKbdSelect->redraw();
    }
}

void restoreRunBrowser() {
    // Read the history of commands in runBrowser on start
    runBrowser->clear();
    Fl_String historyString;
    pGlobalConfig.get("Panel", "RunHistory", historyString,"");
    Fl_String_List history(historyString,"|");
    for (unsigned i = 0; i < history.count(); i++) {
        runBrowser->add(history[i]);
    }
}

void CB_setKeyboard(Fl_Item *item, void *pData)
{
    if (!pGlobalConfig.error() && item) {
        pGlobalConfig.set("Keyboard", "Layout", item->label());
        pGlobalConfig.flush();
        setKeyboard(item->label());
    }
}

void addKeyboards(PanelButton *mPanelButton)
{
    char *countries[49] = {
    "us",   "en_US",  "us_intl",  "am", "az", "by",  "be",  "br",		
    "bg",  "ca",    "cz",  "cz_qwerty",	 "dk",	 "dvorak",  "ee",		
    "fi",  "fr",  "fr_CH",  "de",  "de_CH",  "el",  "hr",  "hu",
    "is",  "il",  "it",	  "jp",	  "lt",	  "lt_std",  "lt_p",  "lv",
    "mk",  "no",  "pl",	  "pt",	  "ro",	  "ru",	  "sr",	  "si",
    "sk",  "sk_qwerty",  "es",  "se",  "th",  "ua",  "gb",  "vn",
    "nec/jp",  "tr"
    };
  
    for (int i=0; i<49; i++)
    {
        mPanelButton->begin();
        Fl_Item *mKbdItem = new Fl_Item(countries[i]);
        mKbdItem->callback((Fl_Callback*)CB_setKeyboard);
        mKbdItem->image(keyboard_pix);
        mPanelButton->end();
    }
}

void clockRefresh(void *)
{
    mClockBox->label(Fl_Date_Time::Now().time_string().sub_str(0, 5));

    Fl_String pClockTooltip = Fl_Date_Time::Now().day_name() + ", ";
    pClockTooltip += Fl_Date_Time::Now().date_string() + ", " + Fl_Date_Time::Now().time_string();
    mClockBox->tooltip(pClockTooltip);

    mClockBox->redraw();
    Fl::add_timeout(60.0, clockRefresh);
}

void runUtility(Fl_Widget *, char *pCommand)
{
    char pRun[256];
    snprintf(pRun, sizeof(pRun)-1, "%s &", pCommand);
    fl_start_child_process(pRun);
}

void updateSetup()
{
    //printf("updateSetup()\n");
    struct stat s;

    // Check if configuration needs to be updated
    if(!stat(pGlobalConfig.filename(), &s)) {
        static long last_modified = 0;
        long now = s.st_mtime;

        if(last_modified>0 && last_modified==now) {
            // Return if not modified
            return;
        }
        // Store last modified time
        last_modified = s.st_mtime;
    }

    pGlobalConfig.read_file(false);
    if(pGlobalConfig.error()) {
        fprintf(stderr, "[EWorkPanel Error]: %s\n", pGlobalConfig.strerror());
        return;
    }

    bool auto_hide = false;
    static bool hiden = false;
    static bool last_state = false;
    static bool on_start = true;

    pGlobalConfig.get("Panel", "AutoHide", auto_hide, false);
    if (on_start) {
        last_state = auto_hide;
        restoreRunBrowser() ;
    }

    bool old_hiden = hiden;
    if (auto_hide) {
        mPanelWindow->setAutoHide(1);
        hiden = true;
    } else {
        mPanelWindow->setAutoHide(0);
        hiden = false;
    }

    if(old_hiden!=hiden || on_start) {
        if(!hiden) {
            mPanelWindow->position(mPanelWindow->initX, mPanelWindow->initY);
            if(mPanelWindow->shown()) Fl_WM::set_window_strut(fl_xid(mPanelWindow), 0, 0, 0, mPanelWindow->h());
        } else {
            mPanelWindow->position(mPanelWindow->initX, mPanelWindow->initY+mPanelWindow->h()-4);
            if(mPanelWindow->shown()) Fl_WM::set_window_strut(fl_xid(mPanelWindow), 0, 0, 0, 4);
        }
    }

    on_start = false;
}

void updateStats(void *)
{
    char pLogBuffer[1024];
    static int log_len;
    static struct timeval last_time;
    static long last_rx=0; static long last_tx=0;  static long connect_time=0;
    char *ptr; int fd; char buf[1024];
    long rx = -1; long tx = -1;
    float spi = 0.0; float spo = 0.0;
    static bool modleds = false;
    int len = -1;

    fd = open("/proc/net/dev", O_RDONLY);
    if(fd > 0) {

        len = read(fd, buf, 1023);
        close(fd);

        if(len>0) {
            buf[len] = '\0';
            ptr = strstr( buf, "ppp0" );
        }
    }

    if(fd==-1 || len < 0 || ptr == NULL)
    {
        if (modleds) {
            dock->remove_from_tray(mModemLeds);
            modleds = false;
        }
        last_rx=0; last_tx=0; connect_time=0;

    }
    else
    {
        long dt; int ct; struct timeval tv;

        gettimeofday(&tv, NULL);
	     sscanf(ptr, "%*[^:]:%ld %*d %*d %*d %*d %*d %*d %*d %ld", &rx, &tx); 
	     dt = (tv.tv_sec - last_time.tv_sec) * 1000;
	     dt += (tv.tv_usec - last_time.tv_usec) / 1000;
	     spi = (rx - last_rx) / dt;
	     spi = spi / 1024.0 * 1000.0;
	     spo = (tx - last_tx) / dt;
	     spo = spo / 1024.0 * 1000.0;

	     if ( connect_time == 0 ) 
	         connect_time = tv.tv_sec;
	     ct = (int)(tv.tv_sec - connect_time);

	     snprintf(pLogBuffer, 1024,
			     _("Received: %ld kB (%.1f kB/s)\n"
			     "Sent:     %ld kB (%.1f  kB/s)\n"
			     "Duration: %d min %d sec"), 
			      rx / 1024, spi, tx / 1024, spo, ct / 60, ct % 60 );
	     last_rx = rx;
	     last_tx = tx;
	     last_time.tv_sec = tv.tv_sec;
	     last_time.tv_usec = tv.tv_usec;

	     log_len = 0;

        if (!modleds) 
	     {
	         dock->add_to_tray(mModemLeds);
	         modleds = true;
	     }    

	     if ((int)(spi) > 0) 
	     {
	         mLedIn->color( (Fl_Color)2 );
	         mLedIn->redraw();
	     }    
	     else 
	     {
	         mLedIn->color( (Fl_Color)968701184 );
	         mLedIn->redraw();
	     }

	     if ( (int)(spo) > 0 ) {
	         mLedOut->color( (Fl_Color)2 );
	         mLedOut->redraw();
	     } else {
	         mLedOut->color( (Fl_Color)968701184 );
	         mLedOut->redraw();
        }
        mModemLeds->tooltip(pLogBuffer);
    }

    updateSetup();
    Fl::repeat_timeout(1.0f, updateStats);
}


void getKeyboard(PanelButton *mButton)
{
    Fl_String pKbdLayout;
    pGlobalConfig.get("Keyboard", "Layout", pKbdLayout, "us");
    setKeyboard(pKbdLayout);
}

// Start utility, like "time/date" or "volume"
void startUtility(Fl_Button *, void *pName)
{
    Fl_String value;
    pGlobalConfig.get("Panel", (char*)pName, value, "");

    if(!pGlobalConfig.error() && !value.empty()) {
        value += " &";
        fl_start_child_process(value);
    }
}

void updateWorkspaces(Fl_Widget*,void*)
{
    mWorkspace->clear();
    mWorkspace->begin();

    char **names=0;
    int count = Fl_WM::get_workspace_count();
    int names_count = Fl_WM::get_workspace_names(names);
    int current = Fl_WM::get_current_workspace();

    for(int n=0; n<count; n++) {
        Fl_Item *i = new Fl_Item();
        i->callback( (Fl_Callback *) setWorkspace, (void*)n);
        i->type(Fl_Item::RADIO);
        if(n<names_count && names[n]) {
            i->label(names[n]);
            delete []names[n];
        } else {
            Fl_String tmp;
            i->label(tmp.printf(tmp, "%s %d", _("Workspace") ,n+1));
        }
        if(current==n) i->set_value();
    }
    if(names) delete []names;

    mWorkspace->end();
}

void FL_WM_handler(Fl_Widget *w, void *d)
{
    if(Fl_WM::action()==Fl_WM::WINDOW_NAME || Fl_WM::action()==Fl_WM::WINDOW_ICONNAME) {
        tasks->update_name(Fl_WM::window());
    }
    if(Fl_WM::action()==Fl_WM::WINDOW_ACTIVE) {
        tasks->update_active(Fl_WM::get_active_window());
    }
    else if(Fl_WM::action() >= Fl_WM::WINDOW_LIST && Fl_WM::action() <= Fl_WM::WINDOW_DESKTOP) {
        tasks->update();
    }
    else if(Fl_WM::action()>0 && Fl_WM::action()<Fl_WM::DESKTOP_WORKAREA) {
        updateWorkspaces(w, d);
        tasks->update();
    }
}

//#define DEBUG
void terminationHandler(int signum)
{
#ifndef DEBUG
    // to crash the whole X session can make people worried so try it
    // to save anymore...
    execlp("eworkpanel", "eworkpanel", 0);
#else
    fprintf(stderr, "\n\nEWorkPanel has just crashed! Please report bug to efltk-bugs@fltk.net\n\n");
    abort();
#endif        
}

void cb_run_app(Fl_Input *i, void*)
{
    Fl_String exec(i->value());
    exec += " &";
    fl_start_child_process(exec);
    
    Fl_Input_Browser *ib = (Fl_Input_Browser *)i->parent();
    if (!ib->find(i->value())) {
        ib->add(i->value());
        Fl_String_List history;
        unsigned c = 0;
        if (ib->children() > 15) c = ib->children() - 15;
        for (; c < ib->children(); c++) {
             history.append(ib->child(c)->label());
        }
        pGlobalConfig.set("Panel", "RunHistory", history.to_string("|"));
        pGlobalConfig.flush();
    }

    i->value("");
}
void cb_run_app2(Fl_Input_Browser *i, void*) { cb_run_app(i->input(), 0); }

int main(int argc, char **argv)
{
    signal(SIGCHLD, SIG_IGN);
    signal(SIGSEGV, terminationHandler);
    fl_init_locale_support("eworkpanel", PREFIX"/share/locale");
    fl_init_images_lib();

    int X=0,Y=0,W=Fl::w(),H=Fl::h();

    // Get current workarea
    Fl_WM::get_workarea(X,Y,W,H);

    //printf("Free area: %d %d %d %d\n", X,Y,W,H);

    // We expect that other docks are moving away from panel :)
    //    mPanelWindow = new Fl_Update_Window(X, Fl::h()-28, W, 28, "Workpanel");
    mPanelWindow = new Fl_Update_Window(X, Y+H-30, W, 30, "Workpanel");
    mPanelWindow->layout_spacing(0);
    // Panel is type DOCK
    mPanelWindow->window_type(Fl_WM::DOCK);

    // NOTICE! DO NOT SET OVERRIDE!
    // If override mode, modal windows, such as popup windows
    // wont get any events, if not inside app window.
    // i.e. If You clicked over the menu, it wont disappear at all...
    // USE Fl_WM module window type (NET-WM SPEC)
    //mPanelWindow->clear_border();
    //mPanelWindow->set_override();

    mPanelWindow->setAutoHide(0);

    // Group that holds everything..
    Fl_Group *g = new Fl_Group(0,0,0,0);
    g->box(FL_DIV_UP_BOX);
    g->layout_spacing(2);
    g->layout_align(FL_ALIGN_CLIENT);

    MainMenu *mSystemMenu = new MainMenu();

    Fl_VertDivider *v = new Fl_VertDivider(0, 0, 5, 18, "");
    v->layout_align(FL_ALIGN_LEFT);

    mWorkspace = new PanelButton(0, 0, 26, 22, FL_NO_BOX, FL_DOWN_BOX, "WSMenu");
    mWorkspace->layout_align(FL_ALIGN_LEFT);
    mWorkspace->label_type(FL_NO_LABEL);
    mWorkspace->align(FL_ALIGN_INSIDE|FL_ALIGN_CENTER);
    mWorkspace->image(desktop_pix);
    mWorkspace->tooltip(_("Workspaces"));
    mWorkspace->end();

    v = new Fl_VertDivider(0, 0, 5, 18, "");
    v->layout_align(FL_ALIGN_LEFT);

    //////////////////////////////
    // ALEXEYS PLAYGROUND BEGIN //
    //////////////////////////////

    runBrowser = new Fl_Input_Browser("",100,FL_ALIGN_LEFT,30);
    //runBrowser->image(run_pix);
	//runBrowser->box(FL_THIN_DOWN_BOX);

    // Added _ALWAYS so callback is in case:
    // 1) select old command from input browser
    // 2) press enter to execute. (this won't work w/o _ALWAYS)
    runBrowser->input()->when(FL_WHEN_ENTER_KEY_ALWAYS | FL_WHEN_RELEASE_ALWAYS); 
    runBrowser->input()->callback((Fl_Callback*)cb_run_app);
    runBrowser->callback((Fl_Callback*)cb_run_app2);

    v = new Fl_VertDivider(0, 0, 5, 18, "");
    v->layout_align(FL_ALIGN_LEFT);

    ////////////////////////////
    // ALEXEYS PLAYGROUND END //
    ////////////////////////////

    Fl_Menu_Button *mPopupPanelProp = new Fl_Menu_Button( 0, 0, W, 28 );
    mPopupPanelProp->type( Fl_Menu_Button::POPUP3 );
    mPopupPanelProp->anim_flags(Fl_Menu_::LEFT_TO_RIGHT);
    mPopupPanelProp->anim_speed(0.8);
    mPopupPanelProp->begin();

    /*EItem *mPanelSettings = new EItem(_("Settings"));
    mPanelSettings->x_offset(12);
    mPanelSettings->callback( (Fl_Callback*)runUtility, (void*)"epanelconf" );

    EItem *mAboutItem = new EItem(_("About EDE..."));
    mAboutItem->x_offset(12);
    mAboutItem->callback( (Fl_Callback *)AboutDialog );
*/
    mPopupPanelProp->end();

    tasks = new TaskBar();
    dock = new Dock();

    v = new Fl_VertDivider(0, 0, 5, 18, "");
    v->layout_align(FL_ALIGN_RIGHT);

    {
        // MODEM
        mModemLeds = new Fl_Group(0, 0, 25, 18);
        mModemLeds->box(FL_FLAT_BOX);
        mModemLeds->hide();
        mLedIn = new Fl_Box(2, 5, 10, 10);
        mLedIn->box( FL_OVAL_BOX );
        mLedIn->color( (Fl_Color)968701184);
        mLedOut = new Fl_Box(12, 5, 10, 10);
        mLedOut->box( FL_OVAL_BOX);
        mLedOut->color( (Fl_Color)968701184);
        mModemLeds->end();
    }

    {
        // KEYBOARD SELECT
        mKbdSelect = new PanelButton(0, 0, 20, 18, FL_NO_BOX, FL_DOWN_BOX, "us");
        mKbdSelect->hide();
        mKbdSelect->anim_speed(4);
        mKbdSelect->label_font(mKbdSelect->label_font()->bold());
        mKbdSelect->highlight_color(mKbdSelect->selection_color());
        mKbdSelect->highlight_label_color( mKbdSelect->selection_text_color());

        mKbdSelect->begin();
        addKeyboards(mKbdSelect);
        getKeyboard(mKbdSelect);
        mKbdSelect->end();
    }

    Fl_Button *mSoundMixer;
    {
        // SOUND
        mSoundMixer = new Fl_Button(0, 0, 20, 18);
        mSoundMixer->hide();
        mSoundMixer->box(FL_NO_BOX);
        mSoundMixer->focus_box(FL_NO_BOX);
        mSoundMixer->image(sound_pix);
        mSoundMixer->tooltip(_("Volume control"));
        mSoundMixer->align(FL_ALIGN_INSIDE);
        mSoundMixer->callback( (Fl_Callback*)startUtility, (void*)"Volume Control" );
    }

    {
        // CLOCK
        mClockBox = new Fl_Button(0, 0, 50, 20);
        mClockBox->align(FL_ALIGN_INSIDE|FL_ALIGN_LEFT);
        mClockBox->hide();
        mClockBox->box(FL_FLAT_BOX);
        mClockBox->callback( (Fl_Callback*)startUtility, (void*)"Time and date");
    }

    CPUMonitor *cpumon;
    {
        // CPU monitor
        cpumon = new CPUMonitor();
        cpumon->hide();
    }

    dock->add_to_tray(mClockBox);
    dock->add_to_tray(mKbdSelect);
    dock->add_to_tray(mSoundMixer);
    dock->add_to_tray(cpumon);

    Fl::focus(mSystemMenu);

    mPanelWindow->end();
    mPanelWindow->show(argc, argv);

    Fl_WM::callback(FL_WM_handler, 0, Fl_WM::DESKTOP_COUNT | 
		    					   Fl_WM::DESKTOP_NAMES |
							   Fl_WM::DESKTOP_CHANGED|
							   Fl_WM::WINDOW_LIST|
							   Fl_WM::WINDOW_DESKTOP|
							   Fl_WM::WINDOW_ACTIVE|
							   Fl_WM::WINDOW_NAME|
							   Fl_WM::WINDOW_ICONNAME);

    updateWorkspaces(0,0);

    Fl::add_timeout(0, clockRefresh);
    Fl::add_timeout(0, updateStats);

    while(mPanelWindow->shown())
        Fl::wait();
}
